module JnMtLib
	def randext(n,max)
	  counter = 0

	  ret = []

	  while counter < n
	    id = rand(max)
	    ret << id
	    counter += 1
	  end
	  ret
	end

	def calc_line_ngram_count(words, n)
    c = words.length-n+1
    (c < 0)? words.length : c
	end

	def get_line_ngrams(words, n, ngram_count)
		ngrams = {}
		words.each_index{|idx|
			break if idx == ngram_count
			subwords = words[idx,n]
			ngrams[subwords] ||= 0
			ngrams[subwords] += 1
		}
		ngrams
	end

	def calc_line_clip_count(tngrams, rngramss)
		clipcount = 0
		tngrams.each{|tngram, tnc|
			rnc = 0
			rngramss.each{|rngrams|
				next if rngrams[tngram] == nil
				rnc = rngrams[tngram] if rnc < rngrams[tngram]
			}
			clipcount += (rnc > tnc)? tnc : rnc
		}
		clipcount
	end

	def count_multi_ngrams(tws, rwss, n)
		tcount = 0
		clipcount = 0

		tws.each_index{|lidx|
			tline = tws[lidx]
			tnum= calc_line_ngram_count(tline, n)
			tcount += tnum
			
			tngrams = get_line_ngrams(tline, n, tnum)

			rngramss = []
			rwss.each{|rws|
				rnum = calc_line_ngram_count(rws[lidx], n)
				rngrams = get_line_ngrams(rws[lidx], n, rnum)

				rngramss << rngrams
			}

			clipcount += calc_line_clip_count(tngrams, rngramss)
		}
		[clipcount, tcount]
	end

	def get_best_match_length(tlen, references)
		ret = nil
		mindist = nil
		references.map{|ref| ref.length}.each{|num|
			dist = (tlen-num).abs
			if mindist == nil or dist < mindist
				mindist = dist
				ret = num
			#elsif dist == mindist and num < ret
			#	ret = num
			end
		}
		ret
	end

	module_function :count_multi_ngrams, :randext, :get_best_match_length, :calc_line_ngram_count, :get_line_ngrams, :calc_line_clip_count
end

class BLEU
	def wrap_array(arr, key)
		return nil if arr == nil
		val = arr[key]
		if val.instance_of?(Array)
			val	
		else
			[val]
		end
	end

	def gen_child_data(key)
		ncs = @ngram_clips.map{|nc| wrap_array(nc, key)	}
		tcs = @tcount.map{|tc| wrap_array(tc, key) }
		bml = wrap_array(@best_match_lengths, key)
		[ncs, tcs, bml, @ngram]
	end

	def initialize(target, references, n=4)
		if target == nil
			@ngram_clips = references[0] #ncs
			@tcount = references[1] #tcs
			@best_match_lengths = references[2]
			@ngram = references[3]
			return
		end

		if target.size != references[0].size and not references.inject(true){|pre, ref| pre and target.size == ref.size}
			throw "Array Size Mismatch!"
		end

		twords = target.map{|tgt| tgt.split }
		rwordss = references.map{|ref| ref.map{|rline| rline.split }}
		@ngram_clips = []
		@tcount = []
		@best_match_lengths = []
		@ngram = n

		(1..n).each{|m|
			@ngram_clips[m] = []
			@tcount[m] = []
			twords.each_index{|lidx|		
				tline = twords[lidx]
				tnum = JnMtLib.calc_line_ngram_count(tline, m)
				tngrams = JnMtLib.get_line_ngrams(tline, m, tnum)

				rngramss = []
				rwordss.each{|rwords|
					rline = rwords[lidx]
					rnum = JnMtLib.calc_line_ngram_count(rline, m)
					rngramss << JnMtLib.get_line_ngrams(rline, m, rnum)
				}

				@ngram_clips[m][lidx] = JnMtLib.calc_line_clip_count(tngrams, rngramss)
				@tcount[m][lidx] = tnum
			}
		}
		twords.each_index{|lidx|
			@best_match_lengths[lidx] = JnMtLib.get_best_match_length(twords[lidx].length, rwordss.map{|rws| rws[lidx]})
		}
	end

	def [](key)
		cdat = gen_child_data(key)
		BLEU.new(nil, cdat)
	end

	def +(bleu)
		throw "ngram size mismatch!" if @ngram != bleu.ngram

		newnc = []
		newtc = []
		(1..@ngram).each{|n|
			newnc[n] = @ngram_clips[n] + bleu.ngram_clips[n]
			newtc[n] = @tcount[n] + bleu.tcount[n]
		}

		BLEU.new(nil, [newnc, newtc, @best_match_lengths + bleu.best_match_lengths, @ngram])
	end

	def calc_BLEU(n=@ngram)
		if n > @ngram
			throw "Error! #{n}gram > #{@ngram}gram"
		end

		nclip = @ngram_clips.map{|nc| 
			if nc == nil
				nil
			else
				nc.inject(0){|pre, c| pre+c}
			end
		}
		tc = @tcount.map{|t|
			if t == nil
				nil
			else
				t.inject(0){|pre, c| pre+c}
			end
		}

		pscore = []
		(1..@ngram).each{|n|
			return 0.0 if nclip[n] == 0
			pscore << nclip[n].to_f / tc[n]
		}

		c = tc[1]
		r = @best_match_lengths.inject(0){|pre, count| pre+count}
    bp = (c > r)? 1.0 : Math.exp(1.0-(r.to_f/c))
		#puts "bp=#{bp}, c=#{c}, r=#{r}"

    prebleu = 0.0
    pscore.each{|ps|
      prebleu += Math.log(ps)/4.0
    }
    bleu = bp * Math.exp(prebleu)
	end
	attr_reader :ngram_clips, :tcount, :best_match_lengths, :ngram
end

